/*
              Librera para el gestor de memoria extendida de 32 bits
              -------------------------------------------------------

      Esta librera provee las siguientes funciones:

        X_inicia_gestor()
        X_fin_gestor()
        X_malloc()
        X_free()
        X_coreleft()
        X_leer()
        X_escribir()
        X_direccion()
        X_tamano()

      Actualiza si es necesario el valor de xmserror

                (C) 1994 Miguel Angel Rodrguez Jdar.

*/

#include <stdlib.h>
#include "xms.h"

void initcpu32 (void);
int modo_virtual (void);

static unsigned long mem_total, mem_libre;
static HXMS hmem;
static unsigned long inicio_mem, direcc_dispo;
static inhA20=0;

enum { LIBRE, OCUPADO, BORRADO };
#define MAX_MCB 200
typedef struct {
               char estado;
               unsigned long direccion;
               unsigned long tamano;
               } MCB;
static MCB *tab_alloc;
static int ind_alloc;

#define estado(i) tab_alloc[i].estado
#define direccion(i) tab_alloc[i].direccion
#define tamano(i) tab_alloc[i].tamano

int X_inicia_gestor (unsigned kbtotales)
    {
    int i;

    /* si la mquina est en modo virtual, salir */
    if (modo_virtual())
       return 0;

    /* si no hay gestor XMS, salir */
    if (!xms_iniciar())
       return 0;

    /* construir la tabla de MCB */
    tab_alloc= (MCB *) malloc(MAX_MCB*sizeof(MCB));
    if (!tab_alloc)
       return 0;
            
    /* calcular la cantidad de memoria a asignar por el gestor */  
    if (kbtotales==0)
       mem_total=mem_libre=xms_disponible()*1024L;
    else
       mem_total=mem_libre=kbtotales*1024L;
     
    /* entrar al modo REALMEM 32 bits si no estamos ya en l */
    initcpu32();

    /* asignar la memoria requerida */
    hmem=xms_asignar(mem_total/1024);
    if (xmserror)
       return 0;

    /* bloquear la memoria asignada para obtener su direccin fsica */
    inicio_mem=direcc_dispo=xms_bloquear(hmem);
    if (xmserror)
       return 0;

    /* si la linea A20 no est habilitada, habiltala */
    if (!xms_estadoA20())
       {
       inhA20=1;
       xms_habilA20();
       }

    /* iniciar la tabla de MCB asignados */
    for (i=0;i<MAX_MCB;i++)
        {
        estado(i)=LIBRE;
        direccion(i)=0L;
        tamano(i)=0L;
        }

    /* primera entrada libre en la tabla de MCB */
    ind_alloc=0;

    return 1;
    }

int X_malloc (unsigned long tamano)
    {
    int i;

    /* forzamos a que sea un nmero par por exceso */
    tamano=(tamano+3)&0xfffffffc;

    /* obtener primera entrada disponible en la tabla de MCB */
    for (i=0;i<ind_alloc;i++)
        if (estado(i)==LIBRE || (estado(i)==BORRADO && tamano(i)>=tamano))
           break;
      
    /* si no hay entradas libres, salir */     
    if (i==MAX_MCB)
       return -1;
       
    /* asignar la memoria */
    switch (estado(i))
           {
           case LIBRE:
                if (mem_libre<tamano)
                   return -1;
                estado(i)=OCUPADO;
                direccion(i)=direcc_dispo;
                tamano(i)=tamano;
                mem_libre-=tamano;
                direcc_dispo+=tamano;
                ind_alloc++;
                return i;

           case BORRADO:
                estado(i)=OCUPADO;
                return i;

           default:
                return -1;
           }
    }

void X_fin_gestor (void)
     {
     free (tab_alloc);
     xms_desbloquear(hmem);
     xms_liberar(hmem);
     if (inhA20)
        xms_deshabilA20();
     }

void X_free (int mcb)
     {
     int i,flag;

     /* si el MCB dado es invlido, salir */
     if (mcb>=ind_alloc)
        return;
     if (estado(mcb)!=OCUPADO)
        return;

     /* si es el ltimo MCB asignado, marcarlo como libre */
     if (mcb==ind_alloc-1)
        {
        estado(mcb)=LIBRE;
        direcc_dispo-=tamano(mcb);
        mem_libre+=tamano(mcb);
        direccion(mcb)=0L;
        tamano(mcb)=0L;
        ind_alloc--;
        }
     else
        /* si no, marcarlo como borrado */
        estado(mcb)=BORRADO;

     /* convertir las MCB borrados al final de la tabla en MCB libres */
     for (i=MAX_MCB,flag=0;i>=0;i--)
         {
         if (estado(i)==BORRADO && flag==0)
            flag=1;
         if (estado(i)==BORRADO && flag==1)
            {
            estado(i)=LIBRE;
            direcc_dispo-=tamano(i);
            mem_libre+=tamano(i);
            direccion(i)=0L;
            tamano(i)=0L;
            ind_alloc--;
            }
         if (estado(i)==OCUPADO)
            break;
         }
     }

unsigned long X_direccion (int mcb)
         {
         return direccion(mcb);
         }

unsigned long X_tamano (int mcb)
         {
         return tamano(mcb);
         }

void X_escribir (char *fuente, int mcb, unsigned long offset, unsigned long numbytes)
     {
     /* nos aseguramos de que el nmero es par por defecto */
     numbytes&=0xfffffffc;
	 offset&=0xfffffffc;

     /* hacemos la escritura en memoria extendida */
     xms_copiar(0,fuente,hmem,(FPUNT)(direccion(mcb)-inicio_mem+offset),numbytes);
     }

void X_leer (int mcb, unsigned long offset, char *destino, unsigned long numbytes)
     {
     /* nos aseguramos de que el nmero es par por defecto */
     numbytes&=0xfffffffc;
     offset&=0xfffffffc;

     /* hacemos la lectura desde memoria extendida */
     xms_copiar(hmem,(FPUNT)(direccion(mcb)-inicio_mem+offset),0,destino,numbytes);
     }

unsigned long X_coreleft (void)
         {
         return mem_libre;
         }